'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var _require = require('bluebird'),
    props = _require.props;

var _require2 = require('lodash'),
    isFunction = _require2.isFunction,
    isObject = _require2.isObject,
    mapValues = _require2.mapValues,
    isArray = _require2.isArray,
    size = _require2.size,
    get = _require2.get,
    assign = _require2.assign,
    fromPairs = _require2.fromPairs,
    keys = _require2.keys,
    map = _require2.map,
    difference = _require2.difference;

var pTry = require('./util/pTry');
var parseResult = require('./util/parseResult');
var isEndResult = require('./util/isEndResult');
var isValidationError = require('./util/isValidationError');

/**
 * Given the fields and the rules, get the names of the fields that are illegal
 * @param fields
 * @param rules
 * @returns string[]
 */
var getIllegalFieldNames = function getIllegalFieldNames(fields, rules) {
    return difference(keys(fields), keys(rules));
};

/**
 * Given the fields and the rules, get the errors object for the illegal fields
 * @param fields
 * @param rules
 * @returns {*}
 */
var evaluateIllegal = function evaluateIllegal(fields, rules) {
    return fromPairs(map(getIllegalFieldNames(fields, rules), function (field) {
        return [field, 'Illegal attribute.'];
    }));
};

/**
 * Validate each field in the object using the given rule
 * @param fields
 * @param rules
 * @param options
 * @returns {Promise}
 */
var validateObject = function validateObject(fields, rules, options) {

    var errors = options.strict ? evaluateIllegal(fields, rules) : {};

    return props(mapValues(rules, function (rule, key) {

        return validate(rule, get(fields, key), fields, options);
    })).then(assign.bind(undefined, errors));
};

/**
 * Validate each item in the array using the given rule
 * @param rule
 * @param array
 * @param attributes
 * @param options
 * @returns {Promise}
 */
var validateArray = function validateArray(rule, array, attributes, options) {

    return props(mapValues(array, function (value) {
        return validate(rule, value, attributes, options);
    }));
};

/**
 * Interpret the return value from a validation rule
 * @param {boolean|undefined|null|string|*} result
 * @param {*} args
 * @returns {*}
 */
var evaluateResult = function evaluateResult(result) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
    }

    result = parseResult(result);

    // If the resulting value is a string or null, return that as the validation result
    if (isEndResult(result)) return result;

    // Else loop the result back into the "validate" method with the return value as the new rule
    return validate.apply(undefined, [result].concat(args));
};

/**
 * "Entrypoint" validator
 * @param rule
 * @param value
 * @param {*|null} [attributes] The "rest" of the values that are being validates
 * @param {Object} [options = {}]
 * @param {*} [options.args] These arguments will be passed as "additional" arguments to the validating rules
 * @param {*} [options.strict] Whether to throw a validation error if the input value contains any fields
 *                              that have no corresponding definition in the schema.
 * @returns {*}
 */
var validate = function validate(rule, value) {
    var attributes = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};


    if (isFunction(rule)) {

        // If the provided rule is a function, invoke that function
        // with the value and the rest of the attributes

        var args = options.args ? options.args : [];

        return pTry(function () {
            return rule.apply(undefined, [value, attributes].concat(_toConsumableArray(args)));
        }).then(function (result) {

            return evaluateResult(result, value, attributes, options);
        }).catch(function (e) {

            // Rules that wish to return an object of errors may do so by throwing
            // a ValidationError Containing the object. This is a necessary "step",
            // since if they were to just resolve to an object, that would be
            // interpreted as a subsequent rule.
            if (isValidationError(e)) return e.data;

            throw e;
        });
    }

    if (isArray(rule)) {

        if (size(rule) !== 1) throw new TypeError('Expected the array-type rule to have a length of 1, got ' + size(rule));

        // The given rule is an array so the value should be an array, too
        if (!isArray(value)) return 'Expected an array';

        return validateArray(rule[0], value, attributes, options);
    }

    if (isObject(rule)) {

        // The given rule is an object and thus represents a data structure, so the value should be an object, too
        if (!isObject(value)) return 'Expected an object';

        return validateObject(value, rule, options);
    }

    throw new TypeError('Expected the validation rule to be a function, Schema, object or array, got ' + (typeof rule === 'undefined' ? 'undefined' : _typeof(rule)));
};

module.exports = validate;